/*
 * Created on 02-jun-2005
 *
 * gvSIG. Sistema de Informacin Geogrfica de la Generalitat Valenciana
 *
 * Copyright (C) 2004 IVER T.I. and Generalitat Valenciana.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 *  Generalitat Valenciana
 *   Conselleria d'Infraestructures i Transport
 *   Av. Blasco Ibez, 50
 *   46010 VALENCIA
 *   SPAIN
 *
 *      +34 963862235
 *   gvsig@gva.es
 *      www.gvsig.gva.es
 *
 *    or
 *
 *   IVER T.I. S.A
 *   Salamanca 50
 *   46005 Valencia
 *   Spain
 *
 *   +34 963163400
 *   dac@iver.es
 */
package com.iver.gvsig.geoprocessing.operations;

import java.util.Vector;

import com.hardcode.gdbms.engine.data.DataSource;
import com.hardcode.gdbms.engine.instruction.IncompatibleTypesException;
import com.hardcode.gdbms.engine.values.BooleanValue;
import com.hardcode.gdbms.engine.values.Value;
import com.iver.cit.gvsig.fmap.DriverException;
import com.iver.cit.gvsig.fmap.core.FPolygon2D;
import com.iver.cit.gvsig.fmap.core.FShape;
import com.iver.cit.gvsig.fmap.core.IGeometry;
import com.iver.cit.gvsig.fmap.core.ShapeFactory;
import com.iver.cit.gvsig.fmap.core.v02.FConverter;
import com.iver.cit.gvsig.fmap.drivers.DriverIOException;
import com.iver.cit.gvsig.fmap.layers.FBitSet;
import com.iver.cit.gvsig.fmap.layers.FLayer;
import com.iver.cit.gvsig.fmap.layers.FLayers;
import com.iver.cit.gvsig.fmap.layers.VectorialAdapter;
import com.iver.cit.gvsig.fmap.layers.layerOperations.AlphanumericData;
import com.iver.cit.gvsig.fmap.layers.layerOperations.Selectable;
import com.iver.cit.gvsig.fmap.layers.layerOperations.SingleLayer;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryCollection;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineSegment;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.LinearRing;

/**
 * @author jmorell
 */
public final class GeoProcessingTopologyOperations {
    public static Geometry[] makeBuffers(Geometry[] geometries, double[] bufferDistances, int quadrantSegments) {
        Geometry[] buffers = new Geometry[geometries.length];
        for (int i=0;i<geometries.length;i++) {
            Geometry geometry = geometries[i];
            buffers[i] = geometry.buffer(bufferDistances[i], quadrantSegments);
        }
        return buffers;
    }
    public static Geometry makeDissolveBuffer(Geometry[] geometries, double bufferDistance, int quadrantSegments) {
        GeometryFactory geometryFactory = new GeometryFactory();
        GeometryCollection geometryCollection = new GeometryCollection(geometries,
                geometryFactory);
        Geometry geometryCollectionBuffer = geometryCollection.buffer(bufferDistance, quadrantSegments);
        return geometryCollectionBuffer;
    }
    public static Geometry makeDissolveBuffer(Geometry[] geometries, double[] bufferDistances, int quadrantSegments) {
        Geometry[] buffers = new Geometry[geometries.length];
        for (int i=0;i<geometries.length;i++) {
            Geometry geometry = geometries[i];
            buffers[i] = geometry.buffer(bufferDistances[i], quadrantSegments);
        }
        GeometryFactory geometryFactory = new GeometryFactory();
        GeometryCollection geometryCollection = new GeometryCollection(buffers,
                geometryFactory);
        // Hago un buffer con d=0 solo para poder devolver un POLYGON y no una GEOMETRYCOLLECTION
        // Molara conseguir el mismo efecto de una manera ms ortodoxa ... investigar ...
        Geometry geometryCollectionBuffer = geometryCollection.buffer(0);
        System.out.println("geometryCollection = " + geometryCollection);
        return geometryCollectionBuffer;
    }
    public static Geometry[] makeClip(Geometry[] geometriesToClip, Geometry[] clippingGeometries) {
        /*IGeometry[] result = new IGeometry[geometriesToClip.length];
        System.out.println("El stack se produce al covertir las geometriesToClip a JTS ...");
        Geometry[] jtsGeometriesToClip = new Geometry[geometriesToClip.length];
        for (int i=0;i<geometriesToClip.length;i++) {
            IGeometry geometryToClip = geometriesToClip[i];
            jtsGeometriesToClip[i] = geometryToClip.toJTSGeometry();
        }
        System.out.println("El stack se produce al covertir las clippingGeometries a JTS ...");
        Geometry[] jtsClippingGeometries = new Geometry[clippingGeometries.length];
        for (int i=0;i<clippingGeometries.length;i++) {
            IGeometry clippingGeometry = clippingGeometries[i];
            jtsClippingGeometries[i] = clippingGeometry.toJTSGeometry();
        }
		Selectable selectable = (Selectable)layerToClip;*/
        Vector geometriesClipped = new Vector();
		FBitSet fBitSet = new FBitSet();
        // Convertimos a JTS
		int l=0;
        for (int i=0;i<geometriesToClip.length;i++) {
            Geometry geometryToClip = geometriesToClip[i];
            int k=0;
            for (int j=0;j<clippingGeometries.length;j++) {
                Geometry clippingGeometry = clippingGeometries[j];
                if (clippingGeometry.intersects(geometryToClip)) {
    			    fBitSet.set(i);
                    geometriesClipped.add(geometryToClip.intersection(clippingGeometry));
                    // Para hacer un clip y no un intersect ...
                    if (k>0) {
                        geometriesClipped.set(l, ((Geometry)geometriesClipped.get(l)).union((Geometry)geometriesClipped.get(l-1)));
                    }
                    k++;
                    l++;
                }
            }
        }
        Geometry[] result = new Geometry[geometriesClipped.size()];
        for (int i=0;i<geometriesClipped.size();i++) {
            result[i] = (Geometry)geometriesClipped.get(i);
        }
        return result;
		//selectable.setSelection(fBitSet);
        //
        /*result = new IGeometry[jtsGeometriesClipped.size()];
        for (int i=0;i<jtsGeometriesClipped.size();i++) {
            FShape shapeResult = (FShape)FConverter.jts_to_java2d((Geometry)jtsGeometriesClipped.get(i));
            System.out.println("El stack se produce al construir la geometra " + i);
            System.out.println("shapeResult.getShapeType() = " + shapeResult.getShapeType());
            result[i] = ShapeFactory.createGeometry(shapeResult);
        }
        return result;*/
    }
    public static IGeometry[] makeIntersect(IGeometry[] geometriesToIntersect, IGeometry[] intersectingGeometries, FLayer layerToIntersect) {
        Vector jtsGeometriesIntersected = new Vector();
        IGeometry[] result = new IGeometry[geometriesToIntersect.length];
        Geometry[] jtsGeometriesToClip = new Geometry[geometriesToIntersect.length];
        for (int i=0;i<geometriesToIntersect.length;i++) {
            IGeometry geometryToClip = geometriesToIntersect[i];
            jtsGeometriesToClip[i] = geometryToClip.toJTSGeometry();
        }
        Geometry[] jtsClippingGeometries = new Geometry[intersectingGeometries.length];
        for (int i=0;i<intersectingGeometries.length;i++) {
            IGeometry clippingGeometry = intersectingGeometries[i];
            jtsClippingGeometries[i] = clippingGeometry.toJTSGeometry();
        }
		Selectable selectable = (Selectable)layerToIntersect;
		FBitSet fBitSet = new FBitSet();
        for (int i=0;i<jtsGeometriesToClip.length;i++) {
            Geometry jtsGeometryToClip = jtsGeometriesToClip[i];
            for (int j=0;j<jtsClippingGeometries.length;j++) {
                Geometry jtsClippingGeometry = jtsClippingGeometries[j];
                if (jtsClippingGeometry.intersects(jtsGeometryToClip)) {
    			    fBitSet.set(i);
                    jtsGeometriesIntersected.add(jtsGeometryToClip.intersection(jtsClippingGeometry));
                }
            }
        }
		selectable.setSelection(fBitSet);
        result = new IGeometry[jtsGeometriesIntersected.size()];
        for (int i=0;i<jtsGeometriesIntersected.size();i++) {
            FShape shapeResult = (FShape)FConverter.jts_to_java2d((Geometry)jtsGeometriesIntersected.get(i));
            result[i] = ShapeFactory.createGeometry(shapeResult);
        }
        return result;
    }
    public static IGeometry[] makeUnion(IGeometry[] geometriesToUnion, IGeometry[] uningGeometries, FLayer layerToUnion) {
        Vector jtsGeometriesUned = new Vector();
        IGeometry[] result = new IGeometry[geometriesToUnion.length];
        Geometry[] jtsGeometriesToUnion = new Geometry[geometriesToUnion.length];
        for (int i=0;i<geometriesToUnion.length;i++) {
            IGeometry geometryToUnion = geometriesToUnion[i];
            jtsGeometriesToUnion[i] = geometryToUnion.toJTSGeometry();
        }
        Geometry[] jtsUningGeometries = new Geometry[uningGeometries.length];
        for (int i=0;i<uningGeometries.length;i++) {
            IGeometry uningGeometry = uningGeometries[i];
            jtsUningGeometries[i] = uningGeometry.toJTSGeometry();
        }
		Selectable selectable = (Selectable)layerToUnion;
		FBitSet fBitSet = new FBitSet();
        for (int i=0;i<jtsGeometriesToUnion.length;i++) {
            Geometry jtsGeometryToUnion = jtsGeometriesToUnion[i];
            for (int j=0;j<jtsUningGeometries.length;j++) {
                Geometry jtsUningGeometry = jtsUningGeometries[j];
                if (jtsUningGeometry.intersects(jtsGeometryToUnion)) {
    			    fBitSet.set(i);
                    jtsGeometriesUned.add(jtsGeometryToUnion.intersection(jtsUningGeometry));
                }
            }
        }
        Vector jtsGeometriesUned2 = new Vector();
        for (int i=0;i<jtsGeometriesUned.size();i++) {
            jtsGeometriesUned2.add(jtsGeometriesUned.get(i));
        }
        int num = jtsGeometriesUned.size();
        for (int i=0;i<jtsGeometriesToUnion.length;i++) {
            Geometry jtsGeometryToUnion = jtsGeometriesToUnion[i];
            boolean intersectsWithAnyGeometry = false;
            for (int j=0;j<num;j++) {
                Geometry jtsGeometryUned = (Geometry)jtsGeometriesUned.get(j);
                if (jtsGeometryToUnion.intersects(jtsGeometryUned)) intersectsWithAnyGeometry = true;
            }
            if (intersectsWithAnyGeometry) {
                //Geometry difference = jtsGeometryToUnion.difference(jtsGeometryUned);
                //jtsGeometriesUned2.add(difference);
            } else {
                jtsGeometriesUned2.add(jtsGeometryToUnion);
            }
        }
        for (int i=0;i<jtsUningGeometries.length;i++) {
            Geometry jtsUningGeometry = jtsUningGeometries[i];
            boolean intersectsWithAnyGeometry = false;
            for (int j=0;j<num;j++) {
                Geometry jtsGeometryUned = (Geometry)jtsGeometriesUned.get(j);
                if (jtsUningGeometry.intersects(jtsGeometryUned)) intersectsWithAnyGeometry = true;
            }
            if (intersectsWithAnyGeometry) {
                //Geometry difference = jtsUningGeometry.difference(jtsGeometryUned);
                //jtsGeometriesUned2.add(difference);
            } else {
                jtsGeometriesUned2.add(jtsUningGeometry);
            }
        }
		selectable.setSelection(fBitSet);
        result = new IGeometry[jtsGeometriesUned2.size()];
        for (int i=0;i<jtsGeometriesUned2.size();i++) {
            FShape shapeResult = (FShape)FConverter.jts_to_java2d((Geometry)jtsGeometriesUned2.get(i));
            result[i] = ShapeFactory.createGeometry(shapeResult);
        }
        return result;
    }
    public static IGeometry[] makeDissolve(FLayer input, String fieldSelectedString) {
		// Conseguimos las geometras
        IGeometry[] geometries = new IGeometry[0];
        Selectable inputSelectable = (Selectable)input;
		SingleLayer inputLayer = (SingleLayer)input;
		VectorialAdapter inputLayerVectorialAdapter = inputLayer.getSource();
		try {
            geometries = new IGeometry[inputLayerVectorialAdapter.getShapeCount()];
            for (int i=0;i<inputLayerVectorialAdapter.getShapeCount();i++) {
                geometries[i] = (IGeometry)inputLayerVectorialAdapter.getShape(i);
            }
        } catch (DriverIOException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        }
        // Convertimos a JTS
        Geometry[] jtsGeometries = new Geometry[geometries.length];
        System.out.println("geometries.length = " + geometries.length);
        for (int i=0;i<geometries.length;i++) {
            IGeometry geometry = geometries[i];
            jtsGeometries[i] = geometry.toJTSGeometry();
        }
        // 
        Vector resultGeometries = new Vector(); 
		AlphanumericData lyr = (AlphanumericData)input;
		DataSource ds;
		try {
            ds = lyr.getRecordset();
            int fieldId = ds.getFieldIndexByName(fieldSelectedString);
            Value[] fieldValues = new Value[(int)ds.getRowCount()];
		    int j = 0;
            for (int i=0;i<ds.getRowCount();i++) {
                fieldValues[i] = ds.getFieldValue(i, fieldId);
            }
            for (int i=0;i<fieldValues.length;i++) {
                Value valuei = fieldValues[i];
                IGeometry geometryi = geometries[i];
                Geometry jtsGeometryi = geometryi.toJTSGeometry();
                for (j=0;j<ds.getRowCount();j++) {
                    if (j!=i) {
                        Value valuej = ds.getFieldValue(j, fieldId);
                        if (((BooleanValue)(valuej.equals(valuei))).getValue()) {
                            IGeometry geometryj = geometries[j];
                            Geometry jtsGeometryj = geometryj.toJTSGeometry();
                            if (jtsGeometryj.intersection(jtsGeometryi) instanceof LineString ||
                                    jtsGeometryj.intersection(jtsGeometryi) instanceof LinearRing) {
                                Geometry jtsUnion = jtsGeometryj.union(jtsGeometryi);
                                FPolygon2D union = (FPolygon2D)FConverter.jts_to_java2d(jtsUnion);
                                resultGeometries.add(ShapeFactory.createGeometry(union));
                            }
                        }
                    }
                }
            }
        } catch (DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (com.hardcode.gdbms.engine.data.driver.DriverException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (IncompatibleTypesException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        IGeometry[] result = new IGeometry[resultGeometries.size()];
        for (int i=0;i<resultGeometries.size();i++) {
            result[i] = (IGeometry)resultGeometries.get(i);
        }
        return result;
    }
    public static IGeometry[] makeMerge(FLayer[] inputLayers, FLayer fixedLayer) {
		int geometryCount = 0;
        for (int i=0;i<inputLayers.length;i++) {
		    VectorialAdapter inputLayerVectorialAdapter = ((SingleLayer)inputLayers[i]).getSource();
		    try {
                geometryCount = geometryCount + inputLayerVectorialAdapter.getShapeCount();
            } catch (DriverIOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
		}
        IGeometry[] geometries = new IGeometry[geometryCount];
        int k=0;
        for (int i=0;i<inputLayers.length;i++) {
		    VectorialAdapter inputLayerVectorialAdapter = ((SingleLayer)inputLayers[i]).getSource();
		    try {
                for (int j=0;j<inputLayerVectorialAdapter.getShapeCount();j++) {
                    geometries[k] = inputLayerVectorialAdapter.getShape(j);
                    k++;
                }
            } catch (DriverIOException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
		}
	    return geometries;
    }
    public static IGeometry[] makeSpatialJoin(FLayer layerToSJ, FLayer SJLayer) {
	    VectorialAdapter inputLayerVectorialAdapter = ((SingleLayer)layerToSJ).getSource();
        int geometryCount=0;
        IGeometry[] geometries = null;
        try {
            geometryCount = inputLayerVectorialAdapter.getShapeCount();
            geometries = new IGeometry[geometryCount];
            for (int j=0;j<inputLayerVectorialAdapter.getShapeCount();j++) {
                geometries[j] = inputLayerVectorialAdapter.getShape(j);
            }
        } catch (DriverIOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return geometries;
    }
	/*private static FPolygon2D points2DToFPolygon2D(Point2D[] pts) {
		GeneralPathX genPathX = new GeneralPathX();
		genPathX.moveTo(pts[0].getX(), pts[0].getY());
		for (int i=1; i<pts.length; i++) {
			genPathX.lineTo(pts[i].getX(), pts[i].getY());
		}
		genPathX.closePath();
		return new FPolygon2D(genPathX);
	}*/
}
